from . import (
    admin,
    block_documents,
    flows,
    run_history,
    flow_runs,
    task_runs,
    flow_run_notification_policies,
    flow_run_states,
    task_run_states,
    deployments,
    saved_searches,
    dependencies,
    logs,
    concurrency_limits,
    workers,
    work_queues,
    artifacts,
    block_schemas,
    block_types,
    block_capabilities,
    collections,
    variables,
    ui,
    root,
    # Server relies on all of the above routes
    server,
)
