import{a as p,x as d,$ as k,c as s,f as a,w as n,g as _,o as t,j as f,u as e,aX as i,i as x,aY as b,aZ as g,F as h,h as y}from"./index-e7fdb3d4.js";import{u as B}from"./usePageTitle-df55e3a5.js";const N=p({__name:"Blocks",setup(D){const l=d(),o=k(l.blockDocuments.getBlockDocuments),c=s(()=>o.response??[]),u=s(()=>o.executed&&c.value.length==0),r=s(()=>o.executed);return B("Blocks"),(C,v)=>{const m=_("p-layout-default");return t(),a(m,{class:"blocks"},{header:n(()=>[f(e(i))]),default:n(()=>[e(r)?(t(),x(h,{key:0},[e(u)?(t(),a(e(b),{key:0})):(t(),a(e(g),{key:1,"block-documents":e(c),onDelete:e(o).refresh},null,8,["block-documents","onDelete"]))],64)):y("",!0)]),_:1})}}});export{N as default};
