import{a as l,r as f,e as u,f as c,w as t,g as _,o as p,j as r,u as o,bq as i,br as m}from"./index-e7fdb3d4.js";import{u as b}from"./usePageTitle-df55e3a5.js";const v=l({__name:"Variables",setup(d){const a=f(),s=()=>{var e;(e=a.value)==null||e.refreshSubscriptions()};return b(u.info.variables),(e,h)=>{const n=_("p-layout-default");return p(),c(n,{class:"variables"},{header:t(()=>[r(o(i),{onCreate:s})]),default:t(()=>[r(o(m),{ref_key:"table",ref:a,onDelete:s},null,512)]),_:1})}}});export{v as default};
