import{c as o,u,bx as a}from"./index-e7fdb3d4.js";function i(...t){const r=[...t,"Prefect Server"],c=o(()=>r.map(e=>u(e)).filter(e=>e!==null).join(" • "));a(()=>document.title=c.value)}export{i as u};
