import{a as D,x as g,b as C,c as s,D as B,$ as I,al as N,am as R,f as l,w as o,g as i,B as $,o as a,u as e,an as V,h as r,ao as w,j as u,ap as j,aq as q,X as z}from"./index-e7fdb3d4.js";import{u as K}from"./usePageTitle-df55e3a5.js";const T=D({__name:"Flow",setup(M){const p=g(),f=C("flowId"),c=s(()=>[f.value]),m=$(),_=s(()=>{const n=["Runs","Deployments"];return B.xl||n.unshift("Details"),n}),d={interval:3e5},b=I(p.flows.getFlow,[f.value],d),t=s(()=>b.response),{filter:y}=N({flows:{id:c}}),{filter:v}=R({flows:{id:c}});function F(){m.push(z.flows())}const h=s(()=>t.value?`Flow: ${t.value.name}`:"Flow");return K(h),(n,O)=>{const k=i("p-tabs"),x=i("p-layout-well");return a(),l(x,{class:"flow"},{header:o(()=>[e(t)?(a(),l(e(V),{key:0,flow:e(t),onDelete:F},null,8,["flow"])):r("",!0)]),well:o(()=>[e(t)?(a(),l(e(w),{key:0,flow:e(t)},null,8,["flow"])):r("",!0)]),default:o(()=>[u(k,{tabs:e(_)},{details:o(()=>[e(t)?(a(),l(e(w),{key:0,flow:e(t)},null,8,["flow"])):r("",!0)]),deployments:o(()=>[u(e(j),{filter:e(v)},null,8,["filter"])]),runs:o(()=>[u(e(q),{"flow-run-filter":e(y)},null,8,["flow-run-filter"])]),_:1},8,["tabs"])]),_:1})}}});export{T as default};
