import{a as d,x as m,$ as _,c as n,f as s,w as c,g,o as a,j as x,u as e,bd as k,i as y,be as N,bf as b,F as v,h as C}from"./index-e7fdb3d4.js";import{u as $}from"./usePageTitle-df55e3a5.js";const S=d({__name:"Notifications",setup(h){const r=m(),t=_(r.notifications.getNotifications),i=n(()=>t.response??[]),f=n(()=>t.executed&&i.value.length===0),u=n(()=>t.executed);return $("Notifications"),(B,o)=>{const l=g("p-layout-default");return a(),s(l,{class:"notifications"},{header:c(()=>[x(e(k))]),default:c(()=>[e(u)?(a(),y(v,{key:0},[e(f)?(a(),s(e(N),{key:0})):(a(),s(e(b),{key:1,notifications:e(i),onDelete:o[0]||(o[0]=p=>e(t).refresh()),onUpdate:o[1]||(o[1]=p=>e(t).refresh())},null,8,["notifications"]))],64)):C("",!0)]),_:1})}}});export{S as default};
