import{a as b,x as d,b as i,c as t,ad as _,u as o,f,w as s,h as B,g as D,B as v,o as g,j as n,b9 as h,ba as x,X as w}from"./index-e7fdb3d4.js";import{u as C}from"./usePageTitle-df55e3a5.js";const T=b({__name:"BlockView",setup(y){const u=d(),a=v(),c=i("blockDocumentId"),l=t(()=>c.value?[c.value]:null),r=_(u.blockDocuments.getBlockDocument,l),e=t(()=>r.response),m=()=>{a.push(w.blocks())},k=t(()=>e.value?`Block: ${e.value.name}`:"Block");return C(k),(N,V)=>{const p=D("p-layout-default");return o(e)?(g(),f(p,{key:0,class:"block-view"},{header:s(()=>[n(o(h),{"block-document":o(e),onDelete:m},null,8,["block-document"])]),default:s(()=>[n(o(x),{"block-document":o(e)},null,8,["block-document"])]),_:1})):B("",!0)}}});export{T as default};
