import{a as C,x as h,b as x,c as o,D as g,T as w,$ as T,f as c,w as n,g as a,B as D,o as r,u as e,bj as B,h as s,bk as u,j as R,bl as S,X as $}from"./index-e7fdb3d4.js";import{u as j}from"./usePageTitle-df55e3a5.js";const O=C({__name:"ConcurrencyLimit",setup(I){const l=h(),m=x("concurrencyLimitId"),y=D(),p=o(()=>[{label:"Details",hidden:g.xl},{label:"Active Task Runs"}]),{tabs:_}=w(p),d={interval:3e5},b=T(l.concurrencyLimits.getConcurrencyLimit,[m.value],d),t=o(()=>b.response);function L(){y.push($.concurrencyLimits())}const v=o(()=>t.value?`Concurrency Limit: ${t.value.tag}`:"Concurrency Limit");return j(v),(N,V)=>{const f=a("p-tabs"),k=a("p-layout-well");return r(),c(k,{class:"concurrencyLimit"},{header:n(()=>[e(t)?(r(),c(e(B),{key:0,"concurrency-limit":e(t),onDelete:L},null,8,["concurrency-limit"])):s("",!0)]),well:n(()=>[e(t)?(r(),c(e(u),{key:0,alternate:"","concurrency-limit":e(t)},null,8,["concurrency-limit"])):s("",!0)]),default:n(()=>[R(f,{tabs:e(_)},{details:n(()=>[e(t)?(r(),c(e(u),{key:0,"concurrency-limit":e(t)},null,8,["concurrency-limit"])):s("",!0)]),"active-task-runs":n(()=>{var i;return[(i=e(t))!=null&&i.activeSlots?(r(),c(e(S),{key:0,"active-slots":e(t).activeSlots},null,8,["active-slots"])):s("",!0)]}),_:1},8,["tabs"])]),_:1})}}});export{O as default};
