import{a as p,bm as i,f as _,w as t,g as a,o as m,j as n,u as o,bn as d,q as f,bo as b,bp as C}from"./index-e7fdb3d4.js";const x=p({__name:"ConcurrencyLimits",setup(k){const{showModal:s,open:l}=i();return(w,e)=>{const c=a("p-button"),u=a("p-layout-default");return m(),_(u,{class:"concurrency-limits"},{header:t(()=>[n(o(d),{crumbs:[{text:"Task Run Concurrency Limits"}]},{"after-crumbs":t(()=>[n(c,{inset:"",size:"xs",icon:"PlusIcon",onClick:o(l)},null,8,["onClick"])]),_:1}),n(o(b),{showModal:o(s),"onUpdate:showModal":e[0]||(e[0]=r=>f(s)?s.value=r:null)},null,8,["showModal"])]),default:t(()=>[n(o(C))]),_:1})}}});export{x as default};
