import{a as _,x as k,f as d,w as a,g as f,B as m,o as w,j as t,u,aS as h,aT as Q,X as s,aB as c}from"./index-e7fdb3d4.js";import{u as C}from"./usePageTitle-df55e3a5.js";const B=_({__name:"WorkQueueCreate",setup(x){const n=k(),o=m(),l=()=>{o.push(s.workQueues())},p=async r=>{try{const{id:e}=await n.workQueues.createWorkQueue(r);c("Work queue has been created","success"),o.push(s.workQueue(e))}catch(e){c("Error occurred while creating new work queue","error"),console.error(e)}};return C("Create Work Queue"),(r,e)=>{const i=f("p-layout-default");return w(),d(i,null,{header:a(()=>[t(u(h))]),default:a(()=>[t(u(Q),{action:"Create",onSubmit:p,onCancel:l})]),_:1})}}});export{B as default};
