import{a as d,b as p,x as R,$ as m,c as n,f as s,w as t,g as v,B as g,o as u,u as e,ab as h,h as x,j as F,ac as y,X as k}from"./index-e7fdb3d4.js";import{u as C}from"./usePageTitle-df55e3a5.js";const I=d({__name:"FlowRunRadar",setup(V){const l=g(),a=p("flowRunId"),r={interval:5e3},c=R(),f=m(c.flowRuns.getFlowRun,[a.value],r),o=n(()=>f.response);function w(){l.push(k.flowRuns())}const _=n(()=>o.value?`Radar View for Flow Run: ${o.value.name}`:"Radar View for Flow Run");return C(_),(b,B)=>{const i=v("p-layout-full");return u(),s(i,{class:"flow-run-radar"},{header:t(()=>[e(o)?(u(),s(e(h),{key:0,"flow-run":e(o),class:"flow-run-radar__header",onDelete:w},null,8,["flow-run"])):x("",!0)]),default:t(()=>[F(e(y),{"flow-run-id":e(a)},null,8,["flow-run-id"])]),_:1})}}});export{I as default};
