import{a as m,x as i,$ as k,c as s,f as a,w as n,g as _,o as t,j as f,u as e,by as x,i as w,bz as y,bA as P,F as b,h}from"./index-e7fdb3d4.js";import{u as g}from"./usePageTitle-df55e3a5.js";const N=m({__name:"WorkPools",setup(C){const r=i(),c={interval:3e4},o=k(r.workPools.getWorkPools,[{}],c),l=s(()=>o.response??[]),p=s(()=>o.executed&&l.value.length==0),u=s(()=>o.executed);return g("Work Pools"),(v,B)=>{const d=_("p-layout-default");return t(),a(d,{class:"work-pools"},{header:n(()=>[f(e(x))]),default:n(()=>[e(u)?(t(),w(b,{key:0},[e(p)?(t(),a(e(y),{key:0})):(t(),a(e(P),{key:1,onUpdate:e(o).refresh},null,8,["onUpdate"]))],64)):h("",!0)]),_:1})}}});export{N as default};
