import{a as k,x as w,b as l,aU as c,f as _,w as r,g as i,o as f,j as n,u as o,bK as P,bL as d}from"./index-e7fdb3d4.js";import{u as x}from"./usePageTitle-df55e3a5.js";const C=k({__name:"WorkPoolQueueEdit",async setup(Q){let e,u;const s=w(),a=l("workPoolName"),t=l("workPoolQueueName"),p=([e,u]=c(()=>s.workPoolQueues.getWorkPoolQueueByName(a.value,t.value)),e=await e,u(),e);return x("Edit Work Pool Queue"),(N,h)=>{const m=i("p-layout-default");return f(),_(m,null,{header:r(()=>[n(o(P),{"work-pool-name":o(a),"work-pool-queue-name":o(t)},null,8,["work-pool-name","work-pool-queue-name"])]),default:r(()=>[n(o(d),{"work-pool-name":o(a),"work-pool-queue":o(p)},null,8,["work-pool-name","work-pool-queue"])]),_:1})}}});export{C as default};
