import{a as m,x as y,b as k,$ as _,c as s,e as o,f as r,w as n,g as d,o as e,u as t,A as g,h as c,k as C,i as x,s as h,F as A}from"./index-e7fdb3d4.js";import{u as $}from"./usePageTitle-df55e3a5.js";const F=m({__name:"ArtifactKey",setup(v){const i=y(),f=k("artifactKey"),u=_(i.artifacts.getArtifactCollection,[f]),a=s(()=>u.response),l=s(()=>a.value?`${o.info.artifact}: ${a.value.key}`:o.info.artifact);return $(l),(B,K)=>{const p=d("p-layout-default");return e(),r(p,{class:"artifact"},{header:n(()=>[t(a)?(e(),r(t(g),{key:0,artifact:t(a)},null,8,["artifact"])):c("",!0)]),default:n(()=>[t(a)?(e(),r(t(C),{key:0,artifact:t(a)},null,8,["artifact"])):c("",!0),t(a)?(e(),x(A,{key:1},[t(a).key?(e(),r(t(h),{key:0,"artifact-key":t(a).key},null,8,["artifact-key"])):c("",!0)],64)):c("",!0)]),_:1})}}});export{F as default};
