import{a as b,c as x,o as i,i as u,j as e,w as t,F as C,bM as h,u as s,bN as w,t as f,bO as B,bP as P,aU as A,f as M,g as o,bn as V,bQ as j,bR as g,q as F}from"./index-e7fdb3d4.js";import{u as N}from"./usePageTitle-df55e3a5.js";const O={class:"settings-block"},R=b({__name:"SettingsCodeBlock",props:{engineSettings:null},setup(p){const n=p,a=x(()=>Object.entries(n.engineSettings));return(d,_)=>(i(),u("div",O,[e(s(w),{multiline:""},{default:t(()=>[(i(!0),u(C,null,h(s(a),(l,c)=>(i(),u("div",{key:c,class:"settings-block--code-line"},f(l[0])+": "+f(l[1]),1))),128))]),_:1})]))}});function T(){return B(P)}const E=b({__name:"Settings",async setup(p){let n,a;const d=[{text:"Settings"}],_=T(),[l,c]=([n,a]=A(()=>Promise.all([_.admin.getSettings(),_.admin.getVersion()])),n=await n,a(),n);return N("Settings"),(U,m)=>{const v=o("p-key-value"),S=o("p-theme-toggle"),r=o("p-label"),y=o("p-layout-default");return i(),M(y,{class:"settings"},{header:t(()=>[e(s(V),{crumbs:d},{actions:t(()=>[e(v,{class:"settings__version",label:"Version",value:s(c),alternate:""},null,8,["value"])]),_:1})]),default:t(()=>[e(r,{label:"Theme"},{default:t(()=>[e(S)]),_:1}),e(r,{label:"Color Mode",class:"settings__color-mode"},{default:t(()=>[e(s(j),{selected:s(g),"onUpdate:selected":m[0]||(m[0]=k=>F(g)?g.value=k:null)},null,8,["selected"])]),_:1}),e(r,{label:"Server Settings"},{default:t(()=>[e(R,{class:"settings__code-block","engine-settings":s(l)},null,8,["engine-settings"])]),_:1})]),_:1})}}});export{E as default};
