import{a as c,x as u,b as _,aU as k,f as m,w as l,g as w,o as f,j as r,u as e,bD as i,bE as d}from"./index-e7fdb3d4.js";import{u as P}from"./usePageTitle-df55e3a5.js";const b=c({__name:"WorkPoolEdit",async setup(x){let o,a;const s=u(),n=_("workPoolName"),t=([o,a]=k(()=>s.workPools.getWorkPoolByName(n.value)),o=await o,a(),o);return P("Create Work Pool"),(y,C)=>{const p=w("p-layout-default");return f(),m(p,null,{header:l(()=>[r(e(i),{"work-pool":e(t)},null,8,["work-pool"])]),default:l(()=>[r(e(d),{"work-pool":e(t)},null,8,["work-pool"])]),_:1})}}});export{b as default};
