import{a as l,b as n,f as s,w as a,g as u,o as p,j as t,u as o,bI as m,bJ as _}from"./index-e7fdb3d4.js";import{u as c}from"./usePageTitle-df55e3a5.js";const C=l({__name:"WorkPoolQueueCreate",setup(f){const e=n("workPoolName");return c("Create Work Pool Queue"),(k,w)=>{const r=u("p-layout-default");return p(),s(r,null,{header:a(()=>[t(o(m),{"work-pool-name":o(e)},null,8,["work-pool-name"])]),default:a(()=>[t(o(_),{"work-pool-name":o(e)},null,8,["work-pool-name"])]),_:1})}}});export{C as default};
