import{a as i,x as m,$ as _,c as s,f as a,w as n,g as d,o,j as w,u as e,ai as k,i as x,aj as C,ak as y,F as h,h as F}from"./index-e7fdb3d4.js";import{u as g}from"./usePageTitle-df55e3a5.js";const N=i({__name:"Flows",setup(v){const c=m(),l={interval:3e4},t=_(c.flows.getFlowsCount,[{}],l),r=s(()=>t.response??0),u=s(()=>t.executed&&r.value===0),p=s(()=>t.executed);return g("Flows"),(B,b)=>{const f=d("p-layout-default");return o(),a(f,{class:"flows"},{header:n(()=>[w(e(k))]),default:n(()=>[e(p)?(o(),x(h,{key:0},[e(u)?(o(),a(e(C),{key:0})):(o(),a(e(y),{key:1,onDelete:e(t).refresh},null,8,["onDelete"]))],64)):F("",!0)]),_:1})}}});export{N as default};
