import{a as k,x as b,b as y,c as t,ad as i,u as o,f as m,w as l,h as _,g as f,o as T,j as s,b1 as d,b2 as g}from"./index-e7fdb3d4.js";import{u as v}from"./usePageTitle-df55e3a5.js";const w=k({__name:"BlocksCatalogView",setup(B){const c=b(),a=y("blockTypeSlug"),n=t(()=>a.value?[a.value]:null),u=i(c.blockTypes.getBlockTypeBySlug,n),e=t(()=>u.response),p=t(()=>e.value?`Block Type: ${e.value.name}`:null);return v(p),(x,C)=>{const r=f("p-layout-default");return o(e)?(T(),m(r,{key:0,class:"blocks-catalog-view"},{header:l(()=>[s(o(d),{"block-type":o(e)},null,8,["block-type"])]),default:l(()=>[s(o(g),{"block-type":o(e)},null,8,["block-type"])]),_:1})):_("",!0)}}});export{w as default};
