import{a as d,x as i,$ as y,c as o,f as a,w as n,g as _,o as s,j as f,u as e,ar as x,i as g,as as h,ap as k,F as D,h as C}from"./index-e7fdb3d4.js";import{u as v}from"./usePageTitle-df55e3a5.js";const N=d({__name:"Deployments",setup(B){const l=i(),c={interval:3e4},t=y(l.deployments.getDeployments,[{}],c),p=o(()=>t.response??[]),r=o(()=>t.executed&&p.value.length===0),m=o(()=>t.executed);return v("Deployments"),(V,b)=>{const u=_("p-layout-default");return s(),a(u,{class:"deployments"},{header:n(()=>[f(e(x))]),default:n(()=>[e(m)?(s(),g(D,{key:0},[e(r)?(s(),a(e(h),{key:0})):(s(),a(e(k),{key:1,onDelete:e(t).refresh},null,8,["onDelete"]))],64)):C("",!0)]),_:1})}}});export{N as default};
