import{a as D,x as I,b as R,$ as h,c as _,r as B,T as N,e as p,m as S,f as o,w as e,g as r,o as s,u as t,C as V,h as i,I as d,i as M,j as n,k as j,D as q,l as m,n as z,t as E,p as b,q as H}from"./index-e7fdb3d4.js";import{u as P}from"./usePageTitle-df55e3a5.js";const U={key:0},L=D({__name:"Artifact",setup(F){const v=I(),y=R("artifactId"),k=h(v.artifacts.getArtifact,[y]),a=_(()=>k.response),c=B(!1),{tab:f,tabs:w}=N([{label:"Artifact"},{label:"Details"},{label:"Raw"}]),x=_(()=>a.value?`${p.info.artifact}: ${a.value.key??S(a.value.type)}`:p.info.artifact);return P(x),(G,l)=>{const C=r("p-divider"),g=r("p-button"),T=r("p-content"),$=r("p-tabs"),A=r("p-layout-well");return s(),o(A,{class:"artifact"},{header:e(()=>[t(a)?(s(),o(t(V),{key:0,artifact:t(a)},null,8,["artifact"])):i("",!0)]),well:e(()=>[t(a)?(s(),o(t(d),{key:0,artifact:t(a),alternate:""},null,8,["artifact"])):i("",!0)]),default:e(()=>[t(a)?(s(),M("section",U,[n(t(j),{artifact:t(a)},null,8,["artifact"]),n(C),t(q).xl?(s(),o(T,{key:0},{default:e(()=>[n(t(m),{artifact:t(a)},null,8,["artifact"]),n(g,{class:"artifact__raw-data-button",inset:"",size:"sm",onClick:l[0]||(l[0]=u=>c.value=!c.value)},{default:e(()=>[z(E(c.value?"Hide":"Show")+" raw data ",1)]),_:1}),c.value?(s(),o(t(b),{key:0,artifact:t(a)},null,8,["artifact"])):i("",!0)]),_:1})):(s(),o($,{key:1,selected:t(f),"onUpdate:selected":l[1]||(l[1]=u=>H(f)?f.value=u:null),tabs:t(w)},{artifact:e(()=>[n(t(m),{artifact:t(a)},null,8,["artifact"])]),details:e(()=>[n(t(d),{artifact:t(a)},null,8,["artifact"])]),raw:e(()=>[n(t(b),{artifact:t(a)},null,8,["artifact"])]),_:1},8,["selected","tabs"]))])):i("",!0)]),_:1})}}});export{L as default};
