.flow-runs__list{display:grid;gap:.5rem}.flow-runs__list-controls{position:sticky;top:0px;z-index:10;display:flex;align-items:center;gap:.5rem;background-color:hsl(var(--background-600) / var(--tw-bg-opacity));--tw-bg-opacity: .9;padding:.75rem .5rem}:is(.dark .flow-runs__list-controls){background-color:hsl(var(--background-400) / var(--tw-bg-opacity));--tw-bg-opacity: .9}.flow-runs__list-controls--right{margin-right:auto;display:flex;align-items:center;gap:.5rem}.flow-runs__chart{height:275px}
