import{a as p,x as l,f as _,w as o,g as m,o as C,j as e,u as n,bg as d,bh as h,aC as i,X as r,aB as g}from"./index-e7fdb3d4.js";import{u as x}from"./usePageTitle-df55e3a5.js";const N=p({__name:"NotificationCreate",setup(y){const c=l();async function s(t){try{await c.notifications.createNotification(t),i.push(r.notifications())}catch(a){g("Error creating notification","error"),console.warn(a)}}function u(){i.push(r.notifications())}return x("Create Notification"),(t,a)=>{const f=m("p-layout-default");return C(),_(f,{class:"notification-create"},{header:o(()=>[e(n(d))]),default:o(()=>[e(n(h),{action:"Create",onSubmit:s,onCancel:u})]),_:1})}}});export{N as default};
