import{a as k,x as f,r as y,c as s,a_ as _,$ as m,f as d,w as t,g,o as C,j as l,u as e,a$ as v,b0 as x}from"./index-e7fdb3d4.js";import{u as B}from"./usePageTitle-df55e3a5.js";const S=k({__name:"BlocksCatalog",setup(T){const c=f(),a=y(null),p=s(()=>a.value?[a.value]:[]),{filter:n}=_({blockSchemas:{blockCapabilities:p}}),i=m(c.blockTypes.getBlockTypes,[n]),u=s(()=>i.response??[]);return B("Blocks Catalog"),($,o)=>{const r=g("p-layout-default");return C(),d(r,{class:"blocks-catalog"},{header:t(()=>[l(e(v))]),default:t(()=>[l(e(x),{capability:a.value,"onUpdate:capability":o[0]||(o[0]=b=>a.value=b),"block-types":e(u)},null,8,["capability","block-types"])]),_:1})}}});export{S as default};
