import{a as k,x as f,$ as _,c as a,f as u,w as r,g as x,o,j as w,u as e,aK as y,i as Q,aL as g,aM as v,F as C,h as $}from"./index-e7fdb3d4.js";import{u as h}from"./usePageTitle-df55e3a5.js";const b=k({__name:"WorkQueues",setup(q){const l=f(),p={interval:3e4},s=_(l.workQueues.getWorkQueues,[{}],p),n=a(()=>s.response??[]),c=a(()=>s.executed&&n.value.length==0),m=a(()=>s.executed);return h("Work Queues"),(B,t)=>{const d=x("p-layout-default");return o(),u(d,{class:"queues"},{header:r(()=>[w(e(y))]),default:r(()=>[e(m)?(o(),Q(C,{key:0},[e(c)?(o(),u(e(g),{key:0})):(o(),u(e(v),{key:1,"work-queues":e(n),onUpdate:t[0]||(t[0]=i=>e(s).refresh()),onDelete:t[1]||(t[1]=i=>e(s).refresh())},null,8,["work-queues"]))],64)):$("",!0)]),_:1})}}});export{b as default};
